#include <math.h>
#include <stdio.h>
#include <stdlib.h>

const double pi = 3.141592653589793;

typedef struct { unsigned char r, g, b; } colour;

void hsv2rgb(float h, float s, float v, float *red, float *grn, float *blu) {
  float i, f, p, q, t, r, g, b;
  if (s == 0) { r = g = b = v; } else {
    h = 6 * (h - floorf(h));
    int ii = i = floorf(h);
    f = h - i;
    p = v * (1 - s);
    q = v * (1 - (s * f));
    t = v * (1 - (s * (1 - f)));
    switch(ii) {
      case 0: r = v; g = t; b = p; break;
      case 1: r = q; g = v; b = p; break;
      case 2: r = p; g = v; b = t; break;
      case 3: r = p; g = q; b = v; break;
      case 4: r = t; g = p; b = v; break;
      default:r = v; g = p; b = q; break;
    }
  }
  *red = r;
  *grn = g;
  *blu = b;
}

void rainbow(int width, int height, int maxiter, int i, int j, const int *counts, const float *trans, unsigned char *pixel) {
  int k0 = height * i + j;
  int kx = height * i + j + height;
  int sx = 1;
  if (i == width - 1) {
    kx -= 2 * height;
    sx = -1;
  }
  int ky = height * i + j + 1;
  int sy = 1;
  if (j == height - 1) {
    ky -= 2;
    sy = -1;
  }
  float r, g, b;
  if (counts[k0] == maxiter) {
    r = g = b = 1.0f;
  } else {
    double n0 = counts[k0] + (1.0 - trans[k0]);
    double nx = counts[kx] + (1.0 - trans[kx]);
    double ny = counts[ky] + (1.0 - trans[ky]);
    double zx = sx * (nx - n0);
    double zy = sy * (ny - n0);
    double zz = 1.0;
    double z = sqrt(zx * zx + zy * zy + zz * zz);
    zx /= z;
    zy /= z;
    zz /= z;
    double hue = atan2(zy, zx) / (2 * pi);
    double sat = 0.5 * zz;
    double val = zz;
    hsv2rgb(hue, sat, val, &r, &g, &b);
  }
  int k = (width * j + i) * 3;
  pixel[k+0] = fminf(fmaxf(255 * r, 0), 255);
  pixel[k+1] = fminf(fmaxf(255 * g, 0), 255);
  pixel[k+2] = fminf(fmaxf(255 * b, 0), 255);
}

int main(int argc, char **argv) {
  (void) argc;
  (void) argv;
  int k = getchar();
  int f = getchar();
  int b = getchar();
  if (k != 'K' || f != 'F' || b != 'B') {
    return 1;
  }
  int width;
  fread(&width, sizeof(int), 1, stdin);
  int height;
  fread(&height, sizeof(int), 1, stdin);
  int *counts = malloc(width * height * sizeof(int));
  fread(counts, width * height * sizeof(int), 1, stdin);
  int iterdiv;
  fread(&iterdiv, sizeof(int), 1, stdin);
  int parts;
  fread(&parts, sizeof(int), 1, stdin);
  colour *keys = malloc(parts * sizeof(colour));
  fread(keys, parts * sizeof(colour), 1, stdin);
  int maxiter;
  fread(&maxiter, sizeof(int), 1, stdin);
  float *trans = malloc(width * height * sizeof(float));
  fread(trans, width * height * sizeof(float), 1, stdin);
  unsigned char *pixels = malloc(width * height * 3);
  #pragma omp parallel for schedule(dynamic)
  for (int j = 0; j < height; ++j) {
    for (int i = 0; i < width; ++i) {
      rainbow(width, height, maxiter, i, j, counts, trans, pixels);
    }
  }
  fprintf(stdout, "P6\n%d %d\n255\n", width, height);
  fwrite(pixels, width * height * 3, 1, stdout);
  free(pixels);
  free(trans);
  free(keys);
  free(counts);
  return 0;
}
